package Net::Ajp13;

our $VERSION = '0.0.3';

use Net::Ajp13::Container;
use Net::Ajp13::Packet;
use Net::Ajp13::Request;
use Net::Ajp13::Response;
use Net::Ajp13::SessionManager;
use Net::Ajp13::Log;
use Socket qw(:all TCP_NODELAY TCP_KEEPALIVE);

use Symbol;
use UNIVERSAL;
use Net::hostent;
use strict;
use threads;
use threads::shared;

use constant AJP13_BUFFER_SIZE => 8192;
use constant AJP13_WS_HEADER   => 0x1234;
use constant AJP13_SW_HEADER   => 0x4142;
use constant AJP13_HEADER_SIZE => 4;

# message types from server to container
use constant AJP13_FORWARD_REQUEST => 2;
use constant AJP13_SHUTDOWN        => 7;
use constant AJP13_PING_REQUEST    => 8;
use constant AJP13_CPING_REQUEST   => 10;

# message types from container to server
use constant AJP13_SEND_BODY_CHUNK   => 3;
use constant AJP13_SEND_HEADERS      => 4;
use constant AJP13_END_RESPONSE      => 5;
use constant AJP13_GET_BODY_CHUNK    => 6;
use constant AJP13_CPONG_REPLY       => 9;


use constant HEADER_NUM => [qw(
    accept accept-charset accept-encoding accept-language
    authorization connection content-type content-length
    cookie cookie2 host pragma referer user-agent
)];


use constant SC_REQ_ACCEPT          => 1;
use constant SC_REQ_ACCEPT_CHARSET  => 2;
use constant SC_REQ_ACCEPT_ENCODING => 3;
use constant SC_REQ_ACCEPT_LANGUAGE => 4;
use constant SC_REQ_AUTHORIZATION   => 5;
use constant SC_REQ_CONNECTION      => 6;
use constant SC_REQ_CONTENT_TYPE    => 7;
use constant SC_REQ_CONTENT_LENGTH  => 8;
use constant SC_REQ_COOKIE          => 9;
use constant SC_REQ_COOKIE2         => 10;
use constant SC_REQ_HOST            => 11;
use constant SC_REQ_PRAGMA          => 12;
use constant SC_REQ_REFERER         => 13;
use constant SC_REQ_USER_AGENT      => 14;

# Integer codes for common (optional) request attribute names
use constant SC_A_CONTEXT       => 1;  # XXX Unused
use constant SC_A_SERVLET_PATH  => 2;  # XXX Unused
use constant SC_A_REMOTE_USER   => 3;
use constant SC_A_AUTH_TYPE     => 4;
use constant SC_A_QUERY_STRING  => 5;
use constant SC_A_JVM_ROUTE     => 6;
use constant SC_A_SSL_CERT      => 7;
use constant SC_A_SSL_CIPHER    => 8;
use constant SC_A_SSL_SESSION   => 9;
use constant SC_A_SSL_KEY_SIZE  => 11;  # ajp14 originally, now in ajp13 with jk 1.2/2.0
use constant SC_A_SECRET        => 12;
# Used for attributes which are not in the list above
use constant SC_A_REQ_ATTRIBUTE => 10;
# Terminates list of attributes
use constant SC_A_ARE_DONE      => 0xFF;


sub new {
    my $class = shift;
    my $conf = shift;
    my %args = @_;
    my $self = bless {
        port => $args{port},
        maxthreads => $args{maxthreads},
        interface  => $args{interface} || 'localhost',
        default_container => $args{container} || Net::Ajp13::Container->new($conf),
        containers => {},
        session => undef,
        config => $conf,
        logger => Net::Ajp13::Log->new(
            $args{log} || *STDERR,
            $conf->get('LogLevel')
        ),
    }, $class;
    return $self;
}

sub logger {
    my $self = shift;
    $self->{logger};
}

my $to_exit : shared = 0;

sub just_die {
    lock($to_exit);
    $to_exit = 1;
}

sub run {
    my $self = shift;

    $self->create_server_socket();
    $self->logger->printf(
        'notice', "Listener: %s(%d/tcp) MaxThreads: %d",
        'localhost', $self->getconf('Listen'), $self->getconf('MaxClients')
    );

    my $log = gensym();
    open $log, '>>', $self->getconf('TransferLog')
        or $self->logger->print('error',
            'TransferLog: cannot open logfile: %s: %s',
            $self->getconf('TransferLog'), $!
        );
    $self->{access_log} = Net::Ajp13::Log->new($log, 'debug', sub {
        shift;
        return join $,, @_, "\n";
    });

    $SIG{HUP} = \&just_die;
    $SIG{INT} = \&just_die;
    $SIG{QUIT} = \&just_die;
    $SIG{TERM} = \&just_die;
   
    $self->{session} = Net::Ajp13::SessionManager->new($self->config);

    $self->init_container();

    my @threads;
    for (1..$self->{maxthreads}) {
        my $t = threads->create('handle_ajp13', $self);
        $t->detach;
        push @threads, $t;
    }
    while (!$to_exit) {sleep(5);}  # .....

    $self->logger->print('notice', 'caught Signal, shutting down');
    for my $t (@threads) {
        #$t->join;
    }

    $self->{session}->destroy;
#    handle_ajp13($self);
    return 0;
}

sub config {
    my $self = shift;
    return $self->{config};
}

sub getconf {
    my $self = shift;
    return $self->{config}->get(@_);
}


sub create_server_socket {
    my $self = shift;

    my $s = gensym();
    socket($s, PF_INET, SOCK_STREAM, getprotobyname('tcp'))
        or die "cannot create socket: $!\n";
    setsockopt($s, SOL_SOCKET, SO_REUSEADDR, 1) or warn "REUSEADDR: $!\n";

    bind($s, sockaddr_in($self->{port}, INADDR_ANY));
    listen($s, SOMAXCONN) or die "cannot listen: $!\n";

    $self->{socket} = $s;
}


sub accept {
    my $self = shift;
    my $addr;

    while (!$to_exit) {
        $addr = accept($self->{connection}, $self->{socket});
        last if $addr;
    }
    return undef if $to_exit;

    select $self->{connection}; $|++; select STDOUT;
    setsockopt($self->{connection}, SOL_SOCKET, SO_LINGER, pack 'l2', 0, 0)
        or warn "SO_LINGER: $!\n";;

    setsockopt($self->{connection}, SOL_SOCKET, TCP_KEEPALIVE, 0)
        or warn "SO_KEEPALIVE: $!\n";

    $self->{addr} = $addr;
    my ($port, $iaddr) = sockaddr_in($addr);
    my $name = gethostbyaddr($iaddr, AF_INET);
    $self->logger->printf('debug',
        'client new connect form "%s"',
        $name->[0]
    );
    $self;
}

sub init_container {
    my $self = shift;
    my $conf = $self->config;

    for my $uri (keys %{$conf->{pathconf}}) {
        my $class = $conf->{pathconf}->{$uri}->{SetHandler};
        next unless $class;
        next if exists $self->{containers}->{$class};

        my $obj;
        eval {
            $obj = $class->new($self->config);
        };
        if ($@) {
            $self->logger->printf('err', 'cannot load Container class: "%s"', $class);
        }
        $obj->init();
        $self->{containers}->{$class} = $obj;
    }
}

sub get_container {
    my $self = shift;
    my $req = shift;

    my $request_dir = $req->uri;
    $request_dir =~ s|/[^/]+$|/|;

    my $class = $self->getconf('SetHandler', $request_dir);
    return $self->{containers}->{$class} if exists $self->{containers}->{$class};

    $self->logger->printf('err', 'cannot loaded Contaner object: "%s"', $class);
    return undef;
}

sub get_default_container {
    my $self = shift;
    $self->{default_container};
}

sub handle_ajp13 {
    my $self = shift;
    my $s = $self->{socket};

    $self->logger->printf('info', 'thread #%d start', threads->tid);
    my $request = Net::Ajp13::Request->new;
    my $response = Net::Ajp13::Response->new;
    my $more_request = 1;

    #for (my $i = 0; $i < 1000; $i++) {
    while (!$to_exit) {
        $self->accept;

        my $count = 1;
        $more_request = 1;
        while (!$to_exit && $more_request) {
            $response->wire_handler($self);
            $self->receive_next_request($request);

            # check auth

            if ($request->status == -2) {
                $response->status($request->status);
                $response->send_error($request, $response);
                $request->recycle;
                $response->recycle;
                last;
            }
            if ($request->status == 999) {
                $response->status($request->status);
                $response->send_error($request, $response);
                $request->recycle;
                $response->recycle;
                next;
            }
            if ($request->status != 200) {
                $response->status($request->status);
                $response->send_error($request, $response);

                $request->recycle;
                $response->recycle;
                last;
            }
 
            # setup servlet request env
            my $container = $self->get_container($request); 
            if (UNIVERSAL::isa($container, 'Net::Ajp13::Container')) {
                $container->invoke($request, $response, $self);
            }
            else {
                $response->status(500);
                $self->get_default_container->send_error($request, $response);
                $self->logger->print('err', 'cannot locate Container class');
                $request->recycle;
                $response->recycle;
                last;
            }


            # finisy response.
            if ($response->status > 400) {
                $self->logger->printf('error',
                    '[clinet %s] [uri %s] %s',
                    $request->remote_addr, $request->uri, $response->status_message
                );
            }
            $self->{access_log}->printf('info',
                '%s - - [%s] "%s %s%s %s" %d %s',
                $request->remote_addr,
                _get_clf_date(),
                $request->method,
                $request->uri, ($request->query_string ? '?'.$request->query_string : ''),
                $request->protocol, $response->status,
                $response->content_length > -1 ? $response->content_length : '-'
            );
            $request->recycle;
            $response->recycle;

            shutdown $self->{connection}, 2;
            #close $self->{connection};

            #last;
            $more_request = 0;
        }
    }
}


sub receive_next_request {
    my $self = shift;
    my $r = shift;
    my $packet = $self->read_message;
    if ($packet == undef) {
        return $r->status(500);
    };

    return $self->handle_message($packet, $r);
}

sub handle_message {
    my $self = shift;
    my $packet = shift;
    my $r = shift;

    my $type = $packet->get_byte();
    if ($type == AJP13_FORWARD_REQUEST) {
        $self->logger->print('debug', "[AJP13_FORWARD_REQUEST]");
        return $self->decode_request($packet, $r);
    }
    elsif ($type == AJP13_CPING_REQUEST) {
        die "recv [AJP13_CPING_REQUEST]\n";

    }
    elsif ($type == AJP13_SHUTDOWN) {
        die "recv [AJP13_SHUTDONW]\n";

    }
    else {
        $self->logger->printf('err', "unknown handler type: %d", $type);
        return $r->status(500);
    }
}


sub send_header {
    my $self = shift;
    my $res = shift;
    my $packet = Net::Ajp13::Packet->new;

    my $l = 0;
    my $body;


    # send header
    $packet->append_int(AJP13_SW_HEADER);
    $packet->append_int(0);  # dummy length;
    $packet->append_byte(AJP13_SEND_HEADERS);
    $packet->append_int($res->status);
    my $msg = Net::Ajp13::Response::HTTP_STATUS_LINE->{$res->status};
    $msg =~ s/^\d{3} //;
    $packet->append_message($msg);

    $packet->append_int(scalar $res->headers);
    for my $name ($res->headers) {
        next if $name eq 'Content-Length';
        $packet->append_message($name);
        $packet->append_message($res->header($name));
    }
    if ($res->content_length >= 0) {
        $packet->append_message('Content-Length');
        $packet->append_message($res->content_length);
    }
    $packet->set_int($packet->{length} - AJP13_HEADER_SIZE, 2); # set packet length

    #$l = send $self->{connection}, $packet->{source}, 0;
    $l = syswrite $self->{connection}, $packet->{source}, $packet->length;

    # $packet->recycle;
}

sub send_body_chunk {
    my $self = shift;
    my $res = shift;
    my $packet = Net::Ajp13::Packet->new;

    $packet->append_int(AJP13_SW_HEADER);
    $packet->append_int(0);  # dummy length;
    $packet->append_byte(AJP13_SEND_BODY_CHUNK);
    $packet->append_message($res->content);
    $packet->set_int($packet->{length} - AJP13_HEADER_SIZE, 2);  # set packet length;

    my $l = syswrite $self->{connection}, $packet->{source}, $packet->length;
}

sub send_end_body_chunk {
    my $self = shift;
    my $packet = Net::Ajp13::Packet->new;

    $packet->append_int(AJP13_SW_HEADER);
    $packet->append_int(0);  # dummy length
    $packet->append_byte(AJP13_END_RESPONSE);
    $packet->append_bool(1);  # reuse connection
    $packet->set_int($packet->{length} - AJP13_HEADER_SIZE, 2);

    my $l = syswrite $self->{connection}, $packet->{source};
}


sub get_body_chunk {
    my $self = shift;
    my $length = shift;

    # create get body chunk msg
    my $send_p = Net::Ajp13::Packet->new();
    $send_p->append_int(AJP13_SW_HEADER);
    $send_p->append_int(0);  # dummy length
    $send_p->append_byte(AJP13_GET_BODY_CHUNK);
    $send_p->append_int($length);

    $send_p->set_int($send_p->{length} - AJP13_HEADER_SIZE, 2);
    # send request
    my $l = syswrite $self->{connection}, $send_p->{source}, $send_p->length;

    # recv body chunk
    my $recv_p = $self->read_message;
    return $recv_p->get_message;
}

sub decode_request {
    my $self = shift;
    my $packet = shift;
    my $r = shift;

    $r->method_num($packet->get_byte());
    $r->protocol($packet->get_message());
    $r->uri($packet->get_message());
    $r->remote_addr($packet->get_message());
    $r->remote_host($packet->get_message());
    $r->server_name($packet->get_message());
    $r->server_port($packet->get_int());
    $r->ssl($packet->get_bool());

    # read http request headers
    my $headers_num = $packet->get_int();
    for (1..$headers_num) {
        my $isc = $packet->peek_int();
        my $id = $isc & 0xFF;
        $isc = $isc & 0xFF00;
        if (0xA000 == $isc) {  # is encoded header
            $packet->get_int();  # skip
            $r->header(HEADER_NUM->[$id - 1], $packet->get_message());

            if ($id == SC_REQ_CONTENT_LENGTH) {
                my $clength = $r->{headers_in}->{'content-length'}
                              ? $r->{headers_in}->{'content-length'}
                              : -1;
                $r->content_length($clength);
            }
            elsif ($id == SC_REQ_CONTENT_TYPE) {
                $r->content_type($r->header('content-type'));
            }
            elsif ($id == SC_REQ_AUTHORIZATION) {
                $r->authorization($r->header('authorization'));
            }
        }
        else {                 # is named header
            my $name = $packet->get_message();
            if ($name == undef) {
                return $r->status(500);
            }
            $r->header($name, $packet->get_message());
        } 
    }


    # read attribute
    for (my $code = $packet->get_byte();
         $code != SC_A_ARE_DONE;
         $code = $packet->get_byte())
    {
        if ($code == SC_A_CONTEXT) { next }
        elsif ($code == SC_A_SERVLET_PATH) { next }
        elsif ($code == SC_A_REMOTE_USER) {
            $r->remote_user($packet->get_message());
        }
        elsif ($code == SC_A_AUTH_TYPE) {
            $r->auth_type($packet->get_message());
        }
        elsif ($code == SC_A_QUERY_STRING) {
            $r->query_string($packet->get_message());
        }
        elsif ($code == SC_A_JVM_ROUTE) {
            $r->jvm_route($packet->get_message());
        }
        elsif ($code == SC_A_SSL_CERT) {
            $r->ssl(1);
            $r->attribute('javax.servlet.request.X509Certificate',
                          $packet->get_message());
        }
        elsif ($code == SC_A_SSL_CIPHER) {
            $r->ssl(1);
            $r->attribute('javax.servlet.request.cipher_suite',
                          $packet->get_message());
        }
        elsif ($code == SC_A_SECRET) {
            $r->secret($packet->get_message());
        }
        elsif ($code == SC_A_SSL_SESSION) {
            $r->attribute('javax.servlet.request.ssl_session',
                          $packet->get_message());
        }
        elsif ($code == SC_A_REQ_ATTRIBUTE) {
            my $name = $packet->get_message();
            $r->attribute($name, $packet->get_message());
        }
        elsif ($code == SC_A_SSL_KEY_SIZE) {
            $r->attribute('javax.servlet.request.key_size', $packet->get_int());
        }
        else {
            $packet->get_message();
        }
    }
    if ($r->ssl) {
        $r->scheme('https');
        $r->secure(1);
    }

    # set cookies on request now that we have all headers

    # receive request body
    if ($r->content_length > 0) {
        my $l = $r->content_length;
        my $rl = 0;

        my $packet = $self->read_message;
        $r->content($packet->get_message());
        $rl += length $r->content;
        while ($r->content_length > $rl) {
            # read more request body
            my $packet = $self->get_body_chunk($r->content_length - $rl);
            my $chunk = $packet->get_message();
            last if $chunk == undef;

            $r->append_content($chunk);
            $rl += length $chunk;
        }
    }

    return $r->status(200);
}

sub read_message {
    my $self = shift;
    my $conn = shift;
    my $addr = shift;

    my $packet = Net::Ajp13::Packet->new();
    my $l = 0;
    while ($l < 4) {
        my $p;
        $l += $self->recv_packet(\$p, 4 - $l);
        $packet->append($p, $l);
    }
    
    if (!$packet->is_valid) {
        $self->logger->printf('crit', 'unknown magic number %X', $packet->magic);
        return undef;
    }

    while ($packet->length > $packet->{src_len} - AJP13_HEADER_SIZE) {
        my $p;
        $l = $self->recv_packet(
            \$p, $packet->length - ($packet->{src_len} - AJP13_HEADER_SIZE)); 
        $packet->append($p, $l);
    }
    return $packet;
}

sub recv_packet {
    my $self = shift;
    my $buffer = shift;
    my $length = shift;
 
    my $l = read $self->{connection}, $$buffer, $length;

    return $l;
}


sub _get_clf_date {
    my $time = time();
    my @t = localtime $time;

    my @month = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);
    sprintf '%02d/%s/%4d:%02d:%02d:%02d',
        $t[3], $month[$t[4]], $t[5]+1900, $t[2], $t[1], $t[0];
}

1;
__END__

=head1 NAME

Net::Ajp13 - mod_jk's ajp13 backhand protocol.

=head1 SYNOPSIS

  use Net::Ajp13;
  use strict;
  
  my $ajp13 = Net::Ajp13->new;
  $ajp13->run();

=head1 DESCRIPTION

hoge.

=head1 SEE ALSO

L<Socket>

=head1 AUTHOR

Hiroyuki OYAMA <oyama@module.jp>

=head1 COPYRIGHT AND LICENSE

Copyright 2004 Hiroyuki OYAMA. Japan. All rights reserved.

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself.

=cut
