package Net::Ajp13::Config;
use strict;

sub new {
    my $class = shift;
    bless {
        path => undef,
        fh => undef,
        ln => 0,
        conf => {},
        pathconf => {},
        _location => '/',
    }, $class;
}

sub to_default {
    my $self = shift;
    my $c = $self->{conf};
    $c->{ServerRoot} = './';
    $c->{TempDir} = './tmp/';
    $c->{Listen} = 8009;
    $c->{MaxClients} = 10;
    $c->{PidFile} =     './logs/atami.pid';
    $c->{TransferLog} = './logs/access_log';
    $c->{ErrorLog} = './logs/error_log';
    $c->{LogLevel} = 'info';
    $c->{OutputEncoding} = undef;
    $c->{MountPath} = {};
    $c->{SetHandler} = 'Net::Ajp13::Container';
}

sub parse {
    my $self = shift;
    my $path = shift;
    require Symbol;

    my $fh = Symbol::gensym();
    open $fh, $path or die "cannot open configuration file: $path: $!\n";
    $self->{path} = $path;
    $self->{fh} = $fh;
    $self->{ln} = 0;
    $self->{conf} = {};

    $self->to_default;
    $self->do_commands();
}

sub do_commands {
    my $self = shift;
    my $fh = $self->{fh};

    while (my $line = <$fh>) {
        $self->{ln}++;
        $line =~ s/#.*//;
        $line =~ s/^[\t\s]+//;
        $line =~ s/\r?\n$//;
        next if $line eq '';

        do {$self->cmd_listen($1); next;}  if $line =~ /^Listen\s+(\d+)$/;
        do {$self->cmd_server_root($1); next;} if $line =~ /^ServerRoot\s+([^\s]+)\s*$/;
        do {$self->cmd_temp_dir($1); next;} if $line =~ /^TempDir\s+([^\s]+)\s*$/;
        do {$self->cmd_max_clients($1); next;} if $line =~ /^MaxClients\s+(\d+)\s*$/;
        do {$self->cmd_pid_file($1); next;} if $line =~ /^PidFile\s+([^\s]+)\s*$/;
        do {$self->cmd_transfer_log($1); next; } if $line =~ /^TransferLog\s+([^\s]+)\s*$/;
        do {$self->cmd_error_log($1); next;} if $line =~ /^ErrorLog\s+([^\s]+)\s*$/;
        do {$self->cmd_log_level($1); next;} if $line =~ /^LogLevel\s+([^\s]+)\s*$/;
        do {$self->cmd_output_encoding($1); next;} if $line =~ /^OutputEncoding\s+([^\s]+)\s*$/;
        do {$self->cmd_mount_path($1, $2); next;} if $line =~ /^MountPath\s+([^\s]+)\s+([^\s]+)\s*$/;
        do {$self->cmd_mount_tar($1, $2); next;} if $line =~ /^MountTar\s+([^\s]+)\s+([^\s]+)\.tar\s*$/;

        do {$self->cmd_set_handler($1); next;} if $line =~ /^SetHandler\s+([^\s]+)\s*$/;
        do {$self->cmd_location($1); next; } if $line =~ /^<Location\s+([^>]+)>\s*$/;
        do {$self->cmd_location_end(); next; } if $line =~ /^<\/Location>\s*$/;

        $self->error('Undefined directive', $line);
    }
}

sub error {
    my $self = shift;
    my $msg = shift;
    my $line = shift;
    die sprintf "%s at %s line %d, \"%s\"\n", $msg, $self->{path}, $self->{ln}, $line;
}

sub cmd_listen {
    my $self = shift;
    my $num = shift;

    if ($self->{_location} ne '/') {
        $self->error('Listen: cannot use in <Location> section');
    }

    if (1 > $num || $num > 65535) {
        $self->error('Listen: Missing, invalid, or non-numeric port', $num);
    }
    $self->{conf}->{Listen} = $num;
}

sub cmd_server_root {
    my $self = shift;
    my $path = shift;
    if ($self->{_location} ne '/') {
        $self->error('ServerRoot: cannot use in <Location> section');
    }

    $path =~ s|/$||;
    $path .= '/';
    if (!-r $path || !-d $path) {
        $self->error('ServerRoot: does not exist or is not a directory', $path);
    }
    $self->{conf}->{ServerRoot} = $path;
}

sub cmd_temp_dir {
    my $self = shift;
    my $path = shift;
    if ($self->{_location} ne '/') {
        $self->error('TempDir: cannot use in <Location> section');
    }

    $path =~ s|/$||;
    $path .= '/';
    if (!-r $path || !-d $path) {
        $self->error('TempDir: does not exist or is not a directory', $path);
    }

    $path = $self->get('ServerRoot'). $path if $path =~ m|^[^/]|;
    $self->{conf}->{TempDir} = $path;
}

sub cmd_max_clients {
    my $self = shift;
    my $num = shift;
    if ($self->{_location} ne '/') {
        $self->error('MaxClients: cannot use in <Location> section');
    }

    if (1 > $num || $num > 65535) {
        $self->error('MaxClients: Require 0 < MaxClientsListen <= 65535', $num);
    }
    $self->{conf}->{MaxClients} = $num;
}

sub cmd_pid_file {
    my $self = shift;
    my $path = shift;
    if ($self->{_location} ne '/') {
        $self->error('TransferLog: cannot use in <Location> section');
    }

    $path = $self->get('ServerRoot'). $path if $path =~ m|^[^/]|;
    $self->{conf}->{PidFile} = $path;
}

sub cmd_transfer_log {
    my $self = shift;
    my $path = shift;
    if ($self->{_location} ne '/') {
        $self->error('TransferLog: cannot use in <Location> section');
    }

    $path = $self->get('ServerRoot'). $path if $path =~ m|^[^/]|;
    $self->{conf}->{TransferLog} = $path;
}

sub cmd_error_log {
    my $self = shift;
    my $path = shift;
    if ($self->{_location} ne '/') {
        $self->error('ErrorLog: cannot use in <Location> section');
    }

    $path = $self->get('ServerRoot'). $path if $path =~ m|^[^/]|;
    $self->{conf}->{ErrorLog} = $path;
}

sub cmd_log_level {
    my $self = shift;
    my $level = shift;

    $self->{conf}->{LogLevel} = lc $level;
}

sub cmd_output_encoding {
    my $self = shift;
    my $enc = shift;

    if ($self->{_location} ne '/') {
        my $uri = $self->{_location};
        $self->{pathconf}->{$uri}->{OutputEncoding} = $enc;
    }
    else {
        $self->{conf}->{OutputEncoding} = $enc;
    }
}

sub cmd_mount_path {
    my $self = shift;
    my $uri = shift;
    my $path = shift;

    if ($self->{_location} ne '/') {
        $self->error('MountPath: cannot use in <Location> section');
    }

    $path = $self->{conf}->{ServerRoot}. $path if $path =~ m|^[^/]|;
    $path =~ s|/?$|/|;
    if (!-r $path || !-d $path) {
        $self->error('MountPath: does not exist or is not a directory', $path);
    }
    $uri =~ s|/?$|/|;
    $self->{conf}->{MountPath}->{$uri} = $path;
}

sub cmd_mount_tar {
    my $self = shift;
    my $uri = shift;
    my $file = shift;

    if ($self->{_location} ne '/') {
        $self->error('MountTar: cannot use in <Location> section');
    }
    $file = $self->{conf}->{ServerRoot}. $file if $file =~ m|^[^/]|;
    if (!-r "$file.tar") {
        $self->error('MountTar: does not exist or is not tar archive file', $file);
    }

    my $dir = $file;
    $dir =~ s|.*/([^/]+)$|$1|;
    $dir = $self->{conf}->{TempDir}. $dir;

    system('tar', '-xf', "$file.tar", '-C', $self->{conf}->{TempDir});
    if (($? >> 8) != 0) {
        $self->error('MountTar: cannot mount tar archive', $file);
        return;
    }
    $uri =~ s|/?$|/|;
    $self->{conf}->{MountPath}->{$uri} = $dir;
}

sub cmd_set_handler {
    my $self = shift;
    my $class = shift;

    eval "require $class";
    if ($@) {
        $self->error("SetHandler: cannot load Container class $class, ", $@);
    }

    if ($self->{_location} ne '/') {
        my $uri = $self->{_location};
        $self->{pathconf}->{$uri}->{SetHandler} = $class;
    }
    else {
        $self->{conf}->{SetHandler} = $class;
    }
}

sub cmd_location {
    my $self = shift;
    my $uri = shift;
    if ($self->{_location} ne '/') {
        $self->error('<Location> without matching <Location>...</Location> section');
    }
    $self->{_location} = $uri;
}
sub cmd_location_end {
    my $self = shift;
    if ($self->{_location} eq '/') {
        $self->error('<Location> without matching <Location>...</Location> section');
    }
    $self->{_location} = '/';
}


sub get {
    my $self = shift;
    my $name = shift;
    my $uri = shift;

    my $value;
    if (defined $uri) {
        $uri =~ s|/$||;
        if (exists $self->{pathconf}->{"$uri/"}) {
            $uri .= '/';
        }
        $value = $self->{pathconf}->{$uri}->{$name};
        if (defined $value) {
            return $value;
        }
    }

    $value = $self->{conf}->{$name};
    if (!defined $value) {
        printf STDERR "[%s] [warn] cannot access conf value: %s\n",
            scalar localtime, $name;
    }
    return $value;
}

1;
__END__
