package Net::Ajp13::Container;
use strict;
use Carp;
use Symbol;

sub new {
    my $class = shift;
    my $conf = shift;

    my $self = bless {
        conf => $conf,
        classes => {},
    }, $class;

    $self->pre_class_load();
    return $self;
}

sub init {
    my $self = shift;

}

sub before {
    my $self = shift;
}

sub invoke {
    my $self = shift;
    my $request = shift;
    my $response = shift;
    my $message_handler = shift;

    my $session = $message_handler->{session}->get_session($request, $response);
    $request->session($session);

    # invoke servlet
    eval {
        my $class = $request->uri;
        my $mount = $message_handler->getconf('MountPath');
        my $path = $class;
        while ($path ne '/') {
            $path =~ s|/[^/]*/?$|/|;
            next unless exists $mount->{$path};
            $class =~ s|^$path|$mount->{$path}|;
            last;
        }

        my $class_path = sprintf '%s.pm', $class;
        die '404' unless -r $class_path;

        my $classname = $self->{classes}->{$class_path};
        $classname->service($request, $response);
    };

    if ($@) {
        if ($@ =~ /^404/) {
            $response->status(404);
        }
        else {
            $response->status(500);
            return $self->send_error($request, $response, $self->_stacktrace());
        }
    }

    if ($response->status >= 300) {
        return $self->send_error($request, $response);
    }

    return $response->close();
}


sub after {
    my $self = shift;
}

sub send_error {
    my $self = shift;
    my $req = shift;
    my $res = shift;
    my $trace = shift;

    return if $res->closed;

    $res->content_type('text/html');

    my $type;
    if ($res->status >= 500) {
        $type = 'Error report';
    }
    else {
        $type = 'Status report';
    }

    my $description;
    if ($res->status == 404) {
        $description = sprintf 'The request resource (%s) is not available.', $req->uri;
    }
    elsif ($res->status == 302) {
        $description = sprintf 'The requested resource (%s) has moved temporarily to a new location.', $req->uri;
    }
    elsif ($res->status == 405) {
        $description = sprintf 'The specified HTTP method is not allowed for the requested resource (%s).', $req->uri;
    }
    elsif ($res->status == 500) {
        $description = sprintf 'The server encountered on internal error (%s) that prevented it from fulfilling this request.', $@;
    }
    else {
        $description = '';
    }

    if ($trace) {
        $trace = sprintf "<p><b>exception</b> <pre>%s%s</pre></p>\n", $@, $trace;
        $type = 'Exception report';
    }

    my $fmt = <<'__HTML__';
<html>
  <head>
    <title>%s/%s - %s</title>
    <style>
<!--
  h1 {
    font-family: Tahoma,Arial,sans-serif;
    color:       white;
    background-color: #525d76;
    font-size: 22px;
  }

  h3 {
    font-family: Tahoma,Arial,sans-serif;
    color: white;
    background-color: #525d76;
    font-size: 14px;
  }

  body {
    font-family: Tahoma,Arial,sans-serif;
    color: black;
    background-color: white;
  }

  b {
    font-family: Tahoma,Arial,sans-serif;
    color: white;
    background-color: #525d76;
  }

  p {
    fint-family: Tahoma,Arial,sans-serif;
    background: white;
    color: black;
    font-size: 12px;
  }

  hr {
    color: #525d76;
  }
-->
    </style>
  </head>
  <body>
    <h1>HTTP %s - %s</h1>
    <hr size="1">
    <p><b>type</b> %s</p>
    <p><b>message</b> <u>%s</u></p>
    <p><b>description</b> <u>%s</u></p>
    %s
    <hr size="1">
    <h3>%s/%s</h3>
  </body>
</html>
__HTML__


    $res->printf($fmt,
        ref $self, $Net::Ajp13::VERSION, $type,
        $res->status_message, $req->uri,
        $type,
        $req->uri,
        $description,
        $trace,
        ref $self, $Net::Ajp13::VERSION
    );
    return $res->close;
}

sub pre_class_load {
    my $self = shift;
    my $c = $self->{conf};
    require File::Find;

    for my $uri (keys %{$c->get('MountPath')}) {
        File::Find::find(sub {
            return unless $File::Find::name =~ /.+\.pm$/;
            my $fh = Symbol::gensym();
            unless (open $fh, $File::Find::name) {
                print STDERR "[%s] [warn] cannot load class: %s: %s\n",
                    scalar localtime, $File::Find::name, $!;
                next;
            }
            while (<$fh>) {
                s/\s*#.*$//;
                if (/^\s*package\s+([^\s;]+);/) {
                    my $pname = $1;
                    $self->{classes}->{$File::Find::name} = $pname;
                    last;
                }
            }
            close $fh;
            do $File::Find::name;
        }, $c->get('MountPath')->{$uri});
    }

}

sub _stacktrace {
    my $self = shift;
    my $i = 0;
    my $str = '';
    while (my @st = caller(++$i)) {
        $str .= sprintf "        %s(%s:%d)\n", $st[3], $st[1], $st[2];
    }
    $str;
}

1;
__END__
